/** -------------------------------------------------
 *                   J-CHOCO
 *   Copyright (C) F. Laburthe, 1999-2003
 * --------------------------------------------------
 *    an open-source Constraint Programming Kernel
 *          for Research and Education
 * --------------------------------------------------
 *
 * file: choco.AbstractConstraint.java
 * last modified by Francois 28 aot 2003:14:59:09
 */
package choco;

import choco.prop.ConstraintPlugin;
import choco.prop.*;
import choco.integer.IntConstraint;
import choco.bool.CompositeConstraint;
import choco.bool.BoolConstraint;
import choco.bool.AbstractCompositeConstraint;

/**
 * An abstract class for all implementations of listeners
 */
public abstract class AbstractConstraint extends AbstractEntity implements Constraint {

  /**
   * The priority of the constraint.
   */

  protected int priority;


  /**
   * The constraint <i>awake</i> var attached to the constraint.
   */

  protected ConstraintEvent constAwakeEvent;


  /**
   * A field for attaching additional data util (useful for PaLM)
   */
  protected ConstraintPlugin hook;


  /**
   * Constraucts a constraint with the priority 0.
   */

  public AbstractConstraint() {
    this(0);
  }


  /**
   * Constructs a constraint with the specified priority.
   * @param priority The wished priority.
   */

  public AbstractConstraint(int priority) {
    this.priority = priority;
    this.constAwakeEvent = new ConstraintEvent(this, false, priority);
  }


  /**
   * Returns the constraint plugin. Useful for extending the solver.
   */

  public ConstraintPlugin getPlugIn() {
    return hook;
  }

  /**
   *  Set the ConstraintPlugin of the constraint. The constraint plugin
   *  gives access to the dependency net of each constraint, it can be useful
   *  to design search heuristic based on explanation.
   * @param hook
   */
  public void setPlugIn(ConstraintPlugin hook) {
    this.hook = hook;
  }

  /**
   * Returns the constraint awake var attached to the constraint.
   */

  public PropagationEvent getEvent() {
    return constAwakeEvent;
  }


  /**
   * Initial propagation of the constraint.
   */

  public void constAwake(boolean isInitialPropagation) {
    ((ChocEngine) getProblem().getPropagationEngine())
        .postConstAwake(this, isInitialPropagation);
  }


  /**
   * Returns the priority.
   */

  public int getPriority() {
    return priority;
  }


  /**
   * Default propagation on variable revision: full constraint re-propagation.
   */

  public void awakeOnVar(int idx) throws ContradictionException {
    propagate();
  }


  /**
   * Default initial propagation: full constraint re-propagation.
   */

  public void awake() throws ContradictionException {
    propagate();
  }


  /**
   * Retrieve the problem of the constraint.
   */

  // ??????????? add the case when the constraint c involves only
  // constant variables, considered as belonging to CURRENT_PB
  public Problem getProblem() {
    int nVars = getNbVars();
    for (int i = 0; i < nVars; i++) {
      Var v = null;
      v = getVar(i);
      if (!(v.getProblem() == null))
        return v.getProblem();
    }
    return null;
  }


  /**
   * Returns the constraint that is in the network and involving the
   * current constraint as a subconstraint.
   */

  Constraint getRootConstraint() {
    Constraint rootConstraint = null;
    rootConstraint = (Constraint) getVar(0).getConstraint(getConstraintIdx(0));
    return rootConstraint;
  }


  /**
   * Un-freezing a constraint (this is useful for mimicking dynamic
   * constraint posts...).
   */

  public void setActive() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      if (!(isActive())) {
        activateListener();
        constAwake(true);
      }
    }
  }


  /**
   * Freezing a constraint (this is useful for backtracking when mimicking
   * dynamic constraint posts...).
   */

  public void setPassive() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      if (isActive()) {
        ConstraintEvent evt = constAwakeEvent;
        EventQueue q = ((ChocEngine) getProblem().getPropagationEngine()).getQueue(evt);
        deactivateListener();
        q.remove(evt);
      }
    }
  }

  /**
   * records that a constraint is now entailed (therefore it is now useless to propagate it again)
   */
  public void setEntailed() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      setPassive();
    } else {
      CompositeConstraint root = (CompositeConstraint) rootConstraint;
      int varOffset = root.getGlobalVarIndex(this, 0);
      ((BoolConstraint) root).setSubConstraintStatus(this, true, varOffset);
    }
  }

  /**
   * Removes a constraint from the network.
   * Beware, this is a permanent removal, it may not be backtracked
   */

  public void delete() {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      setPassive();
      getProblem().eraseConstraint(getSelfIndex());
      for (int i=0; i<getNbVars();i++) {
        AbstractVar v = (AbstractVar) getVar(i);
        int cidx=getConstraintIdx(i);
        v.eraseConstraint(cidx);
      }
    }
  }

  /**
   * returns the (global) index of the constraint among all constraints of the problem
   */
  public int getSelfIndex() {
    return getProblem().getConstraintIndex(this);
  }

  /**
   * making the constraint passive upon events to a variable
   * @param varIndex the index of the variable in the constraint upon which propagation becomes passive
   */
  public void setPassive(int varIndex) {
    Constraint rootConstraint = getRootConstraint();
    if (rootConstraint == this) {
      if (isActive(varIndex)) {
        deactivateListener(varIndex);
      }
    }
  }
  /**
   * raise a contradiction during propagation when the constraint can definitely not be satisfied given the current domains
   */
  public void fail() throws ContradictionException {
    getProblem().getPropagationEngine().raiseContradiction(this);
  }

  public Boolean isEntailed() {
    if (isCompletelyInstantiated()) {
      return Boolean.valueOf(isSatisfied());
    } else {
      return null;
    }

  }

  public boolean isActive(int varIdx) {
      return getVar(varIdx).isActive(getConstraintIdx(varIdx));
    }

    /**
     * Checks if the constraint is active (e.g. plays a role in the propagation phase).
     * @return true if the constraint is indeed currently active
     */

    public boolean isActive() {
      for (int i = 0; i < getNbVars(); i++) {
        if (isActive(i))
          return true;
      }
      return false;
    }

  /**
   * This function connects a constraint with its variables in several ways.
   *  Note that it may only be called once the constraint
   *  has been fully created and is being posted to a problem.
   *  Note that it should be called only once per constraint.
   */

  public void addListener() {
    int n = getNbVars();
    for (int i = 0; i < n; i++) {
      setConstraintIndex(i, getVar(i).addConstraint(this, i, true));
    }
    if (this.hook != null) this.hook.addListener();
  }

  /**
   * Deactivates this listener in all the variables.
   */

  public void deactivateListener() {
    int n = getNbVars();
    for (int i = 0; i < n; i++) {
        getVar(i).deactivateConstraint(this.getConstraintIdx(i));
      }
    if (this.hook != null) this.hook.deactivateListener();
  }

  /**
   * Deactivates this listener in all the variables.
   */

  public void deactivateListener(int varIndex) {
    getVar(varIndex).deactivateConstraint(this.getConstraintIdx(varIndex));
    if (this.hook != null) this.hook.deactivateListener(varIndex);
  }

  /**
   * Activates the listener in all the variables.
   */

  public void activateListener() {
    int n = getNbVars();
    for (int i = 0; i < n; i++) {
        getVar(i).activateConstraint(this.getConstraintIdx(i));
    }
    if (this.hook != null) this.hook.activateListener();
  }

  // defaut implementation: returns the same numbering in a constraint and its counterpart.
  public int getVarIdxInOpposite(int i) {
    return i;
  }

  public AbstractConstraint opposite() {
    throw new UnsupportedOperationException();
  }

  public void connectVar(AbstractVar v, int j) {
    setConstraintIndex(j, v.addConstraint(this, j, true));
  }

  public boolean isEquivalentTo(Constraint compareTo) {
    return this.equals(compareTo);
  }

}
