// **************************************************
// *                   J-CHOCO                      *
// *   Copyright (C) F. Laburthe, 1999-2003         *
// **************************************************
// *  an open-source Constraint Programming Kernel  *
// *     for Research and Education                 *
// **************************************************

/* File choco.bool.AbstractLargeCompositeConstraint.java, last modified by FLABURTHE 17 mars 2004 */
package choco.bool;

import choco.*;
import choco.integer.*;
import choco.integer.var.*;

public abstract class AbstractLargeCompositeConstraint extends AbstractCompositeConstraint {

  /**
   * all subconstraints from the composition
   */
  protected AbstractConstraint[] constraints;

  /**
   * offsets[i] is the total number of variables in all previous terms constraints[0]...constraints[i-1]
   * Therefore it is the offset in the variable indexing of constraints[i]
   */
  protected int[] offsets;

  /**
   * the number of sub-constraints
   */
  protected int nbConstraints;

  /**
   * variables in the composition that are not in the subconstraints
   */
  protected IntDomainVar[] additionalVars;

  /**
   * constraint indices (for the constraint network management) corresponding to the additional variables
   */
  protected int[] additionalIndices;

  public AbstractLargeCompositeConstraint(Constraint[] subConstraints) {
    this(subConstraints, null);
  }

  public AbstractLargeCompositeConstraint(Constraint[] subConstraints, IntVar[] moreVars) {
    int nbc = subConstraints.length;
    constraints = new AbstractConstraint[nbc];
    for (int constIdx = 0; constIdx < nbc; constIdx++) {
      AbstractConstraint c = (AbstractConstraint) subConstraints[constIdx];
      constraints[constIdx] = c;
    }
    nbConstraints = nbc;
    int nbVars = 0;
    offsets = new int[nbc];
    for (int i = 0; i < nbc; i++) {
      nbVars += constraints[i].getNbVars();
      offsets[i] = nbVars;
    }
    int nbv = ((moreVars == null) ? 0 : moreVars.length);
    additionalVars = new IntDomainVar[nbv];
    additionalIndices = new int[nbv];
    for (int j = 0; j < nbv; j++) {
      additionalVars[j] = (IntDomainVar) moreVars[j];
      additionalIndices[j] = 0;
    }
    assert(nbVars + nbv == getNbVars());
  }

  public int assignIndices(AbstractCompositeConstraint root, int i) {
    int j = i;
    for (int constIdx = 0; constIdx < this.nbConstraints; constIdx++) {
      int j0 = j;
      AbstractConstraint subc = constraints[constIdx];
      j = subc.assignIndices(root, j);
      assert(offsets[constIdx] == j - i);
    }
    assert(additionalIndices.length == additionalVars.length);
    for (int varIdx = 0; varIdx < additionalVars.length; varIdx++) {
      j++;
      root.connectVar(additionalVars[varIdx], j);
      this.additionalIndices[varIdx] = additionalVars[varIdx].getNbConstraints();
    }
    return j;
  }

  public int getNbVarsFromSubConst() {
    return offsets[offsets.length - 1];
  }

  public int getSubConstraintIdx(int varIdx) {
    for (int constIdx = 0; constIdx < nbConstraints; constIdx++) {
      if (offsets[constIdx] > varIdx) {
        return constIdx;
      }
    }
    return -1;
  }

  protected int getLocalVarIndex(int varIdx, int constIdx) {
    return ((constIdx == 0) ? varIdx : varIdx - offsets[constIdx - 1]);
  }
  
  public int getVarIdxInOpposite(int i) {
    int constIdx = getSubConstraintIdx(i);
    if (constIdx == 0) {
      return constraints[0].getVarIdxInOpposite(i);
    } else {
      assert(constIdx < constraints.length);
      int off = offsets[constIdx - 1];
      return constraints[constIdx].getVarIdxInOpposite(i - off) + off;
    }
  }

  public void setConstraintIndex(int i, int idx) {
    int constIdx = getSubConstraintIdx(i);
    int varIdx = (constIdx == 0) ? i : i - offsets[constIdx - 1];
    constraints[constIdx].setConstraintIndex(varIdx, idx);
  }

  public int getConstraintIdx(int i) {
    int constIdx = getSubConstraintIdx(i);
    int varIdx = (constIdx == 0) ? i : i - offsets[constIdx - 1];
    return constraints[constIdx].getConstraintIdx(varIdx);
  }

  public int getNbVars() {
    return offsets[nbConstraints - 1] + additionalVars.length;
  }

  public Var getVar(int i) {
    int constIdx = getSubConstraintIdx(i);
    int varIdx = (constIdx == 0) ? i : i - offsets[constIdx - 1];
    return constraints[constIdx].getVar(varIdx);
  }

  public IntVar getIntVar(int varIdx) {
    return (IntVar) getVar(varIdx);
  }

  public boolean isCompletelyInstantiated() {
    boolean instantiated = true;
    for (int constIdx = 0; constIdx < nbConstraints; constIdx++) {
      if (!constraints[constIdx].isCompletelyInstantiated()) {
        instantiated = false;
        break;
      }
    }
    for (int varIdx = 0; varIdx < additionalVars.length; varIdx++) {
      if (!additionalVars[varIdx].isInstantiated()) {
        instantiated = false;
        break;
      }
    }
    return instantiated;
  }

  public AbstractConstraint getSubConstraint(int constIdx) {
    return constraints[constIdx];
  }

  public int getNbSubConstraints() {
    return nbConstraints;
  }
}
