// *********************************************
// *                   J-CHOCO                 *
// *   Copyright (c) F. Laburthe, 1999-2003    *
// *********************************************
// * Event-base contraint programming Engine   *
// *********************************************

// CVS Information
// File:               $RCSfile: NotEqualXYC.java,v $
// Version:            $Revision: 1.1.1.1 $
// Last Modification:  $Date: 2004/05/11 13:51:46 $
// Last Contributor:   $Author: laburthe $

package choco.integer.constraints;

import choco.ContradictionException;
import choco.AbstractConstraint;
import choco.Problem;
import choco.Constraint;
import choco.integer.var.IntDomainVar;
import choco.integer.IntVar;

import java.util.logging.Logger;

//import choco.search.AbstractBinIntConstraint;

/**
 * Implements a constraint X !== Y + C, with X and Y two variables and C a constant.
 */
public class NotEqualXYC extends AbstractBinIntConstraint {

  /**
   * The search constant of the constraint
   */
  protected final int cste;

  /**
   * Constructs the constraint with the specified variables and constant.
   *
   * @param x0 first IntDomainVar
   * @param x1 second IntDomainVar
   * @param c The search constant used in the disequality.
   */

  public NotEqualXYC(choco.integer.IntVar x0, choco.integer.IntVar x1, int c) {
    this.v0 = (choco.integer.var.IntDomainVar) x0;
    this.v1 = (choco.integer.var.IntDomainVar) x1;
    this.cste = c;
  }

  /**
   * The one and only propagation method, using foward checking
   */

  public void propagate() throws ContradictionException {
    if (v0.isInstantiated()) {
      logger.finest("VAL(" + v1.toString() + ") != VAL(" + v0.toString() + ") - " + this.cste);
      v1.removeVal(v0.getValue() - this.cste, this.cIdx1);
    } else if (v1.isInstantiated()) {
      logger.finest("VAL(" + v0.toString() + ") != VAL(" + v1.toString() + ") + " + this.cste);
      v0.removeVal(v1.getValue() + this.cste, this.cIdx0);
    }
  }

  public void awakeOnInf(int idx) throws ContradictionException {
    if (v0.isInstantiated()) {
      logger.finest("VAL(" + v1.toString() + ") != VAL(" + v0.toString() + ") - " + this.cste);
      v1.removeVal(v0.getValue() - this.cste, this.cIdx1);
    } else if (v1.isInstantiated()) {
      logger.finest("VAL(" + v0.toString() + ") != VAL(" + v1.toString() + ") + " + this.cste);
      v0.removeVal(v1.getValue() + this.cste, this.cIdx0);
    }
  }

  public void awakeOnSup(int idx) throws ContradictionException {
    if (v0.isInstantiated()) {
      logger.finest("VAL(" + v1.toString() + ") != VAL(" + v0.toString() + ") - " + this.cste);
      v1.removeVal(v0.getValue() - this.cste, this.cIdx1);
    } else if (v1.isInstantiated()) {
      logger.finest("VAL(" + v0.toString() + ") != VAL(" + v1.toString() + ") + " + this.cste);
      v0.removeVal(v1.getValue() + this.cste, this.cIdx0);
    }
  }

  public void awakeOnInst(int idx) throws ContradictionException {
    if (idx == 0) {
      logger.finest("VAL(" + v1.toString() + ") != VAL(" + v0.toString() + ") - " + this.cste);
      v1.removeVal(v0.getValue() - this.cste, this.cIdx1);
    } else {
      assert(idx == 1);
      logger.finest("VAL(" + v0.toString() + ") != VAL(" + v1.toString() + ") + " + this.cste);
      v0.removeVal(v1.getValue() + this.cste, this.cIdx0);
    }
  }

  public void awakeOnVar(int idx) throws ContradictionException {
    ;
  }

  /**
   * Nothing to do when only a hole in a domain is made
   */

  public void awakeOnRem(int idx, int x) throws ContradictionException {
    ;
  }


  /**
   * Checks if the listeners must be checked or must fail.
   */

  public Boolean isEntailed() {
    if ((v0.getSup() < v1.getInf() + this.cste) ||
        (v1.getSup() < v0.getInf() - this.cste))
      return Boolean.TRUE;
    else if ((v0.getInf() == v0.getSup()) &&
        (v1.getInf() == v1.getSup()) &&
        (v0.getInf() == v1.getInf() + this.cste))
      return Boolean.FALSE;
    else
      return null;
  }

  /**
   * Checks if the constraint is satisfied when the variables are instantiated.
   */

  public boolean isSatisfied() {
    return (v0.getValue() != v1.getValue() + this.cste);
  }

  /**
   * tests if the constraint is consistent with respect to the current state of domains
   * @return true iff the constraint is bound consistent (weaker than arc consistent)
   */
  public boolean isConsistent() {
    return (  (v0.isInstantiated()) ?
                 ( (v1.hasEnumeratedDomain()) ?
                      (!v1.canBeInstantiatedTo(v0.getValue())) :
                      ((v1.getInf() != v0.getValue()) && (v1.getSup() != v0.getValue())) ) :
                 ( (v1.isInstantiated()) ?
                    ( (v0.hasEnumeratedDomain()) ?
                        (!v0.canBeInstantiatedTo(v1.getValue())) :
                        ((v0.getInf() != v1.getValue()) && (v0.getSup() != v1.getValue())) ) :
                    true) );
  }

  public AbstractConstraint opposite() {
//    return new EqualXYC(v0, v1, cste);
    Problem pb = getProblem();
    return (AbstractConstraint) pb.eq(v0, pb.plus(v1, cste));
  }

  public final boolean isEquivalentTo(Constraint compareTo) {
    if (compareTo instanceof NotEqualXYC) {
      NotEqualXYC c = (NotEqualXYC) compareTo;
      return ( ( (this.v0 == c.v0) && (this.v1 == c.v1) && (this.cste == c.cste) )  ||
               ( (this.v0 == c.v1) && (this.v1 == c.v0) && (this.cste == -(c.cste)) ) );
    } else {
      return false;
    }
  }

}
