/** -------------------------------------------------
 *                   J-CHOCO
 *   Copyright (C) F. Laburthe, 1999-2003
 * --------------------------------------------------
 *    an open-source Constraint Programming Kernel
 *          for Research and Education
 * --------------------------------------------------
 *
 * file: choco.mem.Environment.java
 * last modified by Francois 28 aot 2003:14:59:09
 */
package choco.mem;

import choco.util.LightFormatter;

import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.logging.StreamHandler;
import java.security.AccessControlException;

/**
 * The root class for managing memory and sessions.
 * <p/>
 * A environment is associated to each problem.
 * It is responsible for managing backtrackable data.
 */
// new comment
public class Environment {

  /**
   * Index of the {@link choco.mem.StoredIntTrail} for storing integers.
   */

  public final static int INT_TRAIL = 0;


  /**
   * Index of the {@link choco.mem.StoredVectorTrail} for storing vectors.
   */

  public final static int VECTOR_TRAIL = 1;


  /**
   * Index of the {@link choco.mem.StoredIntVectorTrail} for storing
   * search vectors.
   *
   * @see #trails
   */

  public final static int INT_VECTOR_TRAIL = 2;


  /**
   * Index of the {@link choco.mem.StoredFloatTrail} for storing
   * search vectors.
   *
   * @see #trails
   */

  public final static int FLOAT_TRAIL = 3;


  /**
   * The maximum numbers of objects that a
   * {@link choco.mem.ITrailStorage} can handle.
   */

  private int maxHist;


  /**
   * The maximum numbers of worlds that a
   * {@link choco.mem.ITrailStorage} can handle.
   */

  private int maxWorld;


  /**
   * The current world number (should be less
   * than <code>maxWorld</code>).
   */

  public int currentWorld = 0;


  /**
   * Contains all the {@link choco.mem.ITrailStorage} trails for
   * storing different kinds of data.
   */

  protected ITrailStorage[] trails;

  /**
   * Reference to the root Logger object (using the java.util.logging package)
   */

  private static Logger logger = Logger.getLogger("choco");

  /**
   * Reference to the main handler for all choco.* Logger objects
   */

  private static Handler handler = null;

  /**
   * capacity of the trailing stack (in terms of number of worlds that can be handled)
  */
  private int maxWorlds = 0;
  
  /**
   * Constructs a new <code>Environment</code> with
   * the default stack sizes : 50000 and 1000.
   */

  public Environment() {
    maxHist = 50000;
    maxWorld = 1000;
    trails = new ITrailStorage[]{
      new StoredIntTrail(this, maxHist, maxWorld),
      new StoredVectorTrail(this, maxHist, maxWorld),
      new StoredIntVectorTrail(this, maxHist, maxWorld),
      new StoredFloatTrail(this, maxHist, maxWorld)
    };
    try {
      setDefaultHandler();
    } catch(AccessControlException e) {
      // Do nothing if this an applet !
    }
  }

  public void setDefaultHandler() {
    // define default levels, take into account only info, warning and severe messages

    Logger.getLogger("choco").setLevel(Level.ALL);
    Logger.getLogger("choco.search").setLevel(Level.INFO);
    Logger.getLogger("choco.prop").setLevel(Level.SEVERE);
    Logger.getLogger("choco.mem").setLevel(Level.SEVERE);
    Logger.getLogger("choco.integer").setLevel(Level.SEVERE);
    Logger.getLogger("choco.util").setLevel(Level.SEVERE);
    Logger.getLogger("choco.test").setLevel(Level.ALL);
    Logger.getLogger("choco.test.mem").setLevel(Level.SEVERE);
    Logger.getLogger("choco.test.integer").setLevel(Level.SEVERE);
    Logger.getLogger("choco.test.util").setLevel(Level.SEVERE);
    Logger.getLogger("choco.test.search").setLevel(Level.SEVERE);

    StreamHandler sh = new StreamHandler(System.out, new LightFormatter());
    setHandler(sh);
  }

  private void setHandler(Handler h) {
    // remove existing handler on choco logger and define choco handler
    // the handler defined here could be reused by other packages
    Handler[] handlers = logger.getHandlers();
    for (int i = 0; i < handlers.length; i++) {
      logger.removeHandler(handlers[i]);
    }
    // by default, handle (so print) only severe message, it could be modified in other package
    h.setLevel(Level.SEVERE);
    logger.addHandler(h);
    logger.setUseParentHandlers(false);
    handler = h;
  }

  /**
   * Returns the <code>i</code>th trail in the trail array.
   *
   * @param i index of the trail.
   */

  public ITrailStorage getTrail(int i) {
    return trails[i];
  }


  /**
   * Returns the world number.
   *
   * @see #currentWorld
   */

  public int getWorldIndex() {
    return currentWorld;
  }


  /**
   * Starts a new branch in the search tree.
   */

  public void worldPush() {
    for (int i = 0; i < trails.length; i++) {
      ITrailStorage trail = trails[i];
      trail.worldPush();
    }
    currentWorld++;
    if (currentWorld + 1 == maxWorld)
      resizeWorldCapacity(maxWorld * 3 / 2);
  }


  /**
   * Backtracks to the previous choice point in the search tree.
   */
  public void worldPop() {
    for (int i = 0; i < trails.length; i++) {
      ITrailStorage trail = trails[i];
      trail.worldPop();
    }
    currentWorld--;
  }

  /**
   * Accepts all modifications since the previous choice point in the search tree.
   */

  public void worldCommit() {
    for (int i = 0; i < trails.length; i++) {
      ITrailStorage trail = trails[i];
      trail.worldCommit();
    }
    currentWorld--;
  }

  /**
   * Factory pattern: new IStateInt objects are created by the environment
   * (no initial value is assigned to the backtrackable search)
   */

  public IStateInt makeInt() {
    return new StoredInt(this);
  }

  /**
   * Factory pattern: new IStateInt objects are created by the environment
   *
   * @param initialValue the initial value of the backtrackable search
   */

  public IStateInt makeInt(int initialValue) {
    return new StoredInt(this, initialValue);
  }

  /**
   * Factory pattern: new IStateIntVector objects are created by the environment.
   * Creates an empty vector
   */

  public StoredIntVector makeIntVector() {
    return new StoredIntVector(this);
  }

  /**
   * Factory pattern: new IStateIntVector objects are created by the environment
   *
   * @param size         the number of entries in the vector
   * @param initialValue the common initial value for all entries (backtrackable integers)
   */

  public StoredIntVector makeIntVector(int size, int initialValue) {
    return new StoredIntVector(this, size, initialValue);
  }

  /**
   * Factory pattern: new IStateIntVector objects are created by the environment
   *
   * @param entries an array to be copied as set of initial contents of the vector
   */

  public StoredIntVector makeIntVector(int[] entries) {
    return new StoredIntVector(entries);
  }

  /**
   * Factory pattern: new StoredBitSetVector objects are created by the environment
   */
  public StoredBitSet makeBitSet(int size) {
    return new StoredBitSet(this, size);
  }

  public StoredBitSet makeBitSet(int[] entries) {
    return new StoredBitSet(this, 0);  // TODO
  }

  /**
   * Factory pattern: new IStateFloat objects are created by the environment
   * (no initial value is assigned to the backtrackable search)
   */

  public StoredFloat makeFloat() {
    return new StoredFloat(this);
  }

  /**
   * Factory pattern: new IStateFloat objects are created by the environment
   *
   * @param initialValue the initial value of the backtrackable search
   */

  public StoredFloat makeFloat(double  initialValue) {
    return new StoredFloat(this, initialValue);
  }

  /**
   * returns the size of the trail
   */

  public int getTrailSize() {
    int s = 0;
    for (int i = 0; i < trails.length; i++)
      s += trails[i].getSize();
    return s;
  }

  /**
   * returns the size of the trail used for storing integers
   */

  public int getIntTrailSize() {
    return trails[INT_TRAIL].getSize();
  }

  /**
   * returns the size of the trail used for storing search arrays
   */

  public int getIntVectorTrailSize() {
    return trails[INT_VECTOR_TRAIL].getSize();
  }

  private void resizeWorldCapacity(int newWorldCapacity) {
    for (int i = 0; i < trails.length; i++)
      trails[i].resizeWorldCapacity(newWorldCapacity);
    maxWorld = newWorldCapacity;
  }
}

