// *********************************************
// *                   J-CHOCO                 *
// *   Copyright (c) F. Laburthe, 1999-2003    *
// *********************************************
// * Event-base contraint programming Engine   *
// *********************************************

// CVS Information
// File:               $RCSfile: StoredIntVectorTrail.java,v $
// Version:            $Revision: 1.1.1.1 $
// Last Modification:  $Date: 2004/05/11 13:51:47 $
// Last Contributor:   $Author: laburthe $

package choco.mem;

import java.util.logging.Logger;

/**
 * Implements a trail with the history of all the stored search vectors.
 */
public class StoredIntVectorTrail implements ITrailStorage {
  /**
   * Reference to an object for logging trace statements related memory & backtrack (using the java.util.logging package)
   */

  private static Logger logger = Logger.getLogger("choco.mem");

  /**
   * The current environment.
   */

  private Environment environment;


  /**
   * All the stored search vectors.
   */

  private StoredIntVector[] vectorStack;


  /**
   * Indices of the previous values in the stored vectors.
   */

  private int[] indexStack;


  /**
   * Previous values of the stored vector elements.
   */

  private int[] valueStack;


  /**
   * World stamps associated to the previous values
   */

  private int[] stampStack;

  /**
   * The last world an search vector was modified in.
   */

  private int currentLevel;


  /**
   * Starts of levels in all the history arrays.
   */

  private int[] worldStartLevels;

  /**
   * capacity of the trailing stack (in terms of number of updates that can be stored)
   */
  private int maxUpdates = 0;

  /**
   * capacity of the trailing stack (in terms of number of worlds that can be handled)
  */
  private int maxWorlds = 0;

  /**
   * Constructs a trail for the specified environment with the
   * specified numbers of updates and worlds.
   */

  public StoredIntVectorTrail(Environment env, int nUpdates, int nWorlds) {
    this.environment = env;
    this.currentLevel = 0;
    this.vectorStack = new StoredIntVector[nUpdates];
    this.indexStack = new int[nUpdates];
    this.valueStack = new int[nUpdates];
    this.stampStack = new int[nUpdates];
    this.worldStartLevels = new int[nWorlds];
  }


  /**
   * Reacts on the modification of an element in a stored search vector.
   */

  public void savePreviousState(StoredIntVector vect, int index, int oldValue, int oldStamp) {
    this.vectorStack[currentLevel] = vect;
    this.indexStack[currentLevel] = index;
    this.stampStack[currentLevel] = oldStamp;
    this.valueStack[currentLevel] = oldValue;
    currentLevel++;
    if (currentLevel == maxUpdates)
      resizeUpdateCapacity();
}

 private void resizeUpdateCapacity() {
   int newCapacity =  ((maxUpdates * 3) / 2);
   // first, copy the stack of variables
   StoredIntVector[] tmp1 = new StoredIntVector[newCapacity];
   System.arraycopy(vectorStack, 0, tmp1, 0, vectorStack.length);
   vectorStack = tmp1;
   // then, copy the stack of former values
   int[] tmp2 = new int[newCapacity];
   System.arraycopy(valueStack, 0, tmp2, 0, valueStack.length);
   valueStack = tmp2;
   // then, copy the stack of world stamps
   int[] tmp3 = new int[newCapacity];
   System.arraycopy(stampStack, 0, tmp3, 0, stampStack.length);
   stampStack = tmp3;
   // then, copy the stack of indices
   int[] tmp4 = new int[newCapacity];
   System.arraycopy(indexStack, 0, tmp4, 0, indexStack.length);
   indexStack = tmp4;

   // last update the capacity
   maxUpdates = newCapacity;
 }

  public void resizeWorldCapacity(int newWorldCapacity) {
    int[] tmp = new int[newWorldCapacity];
    System.arraycopy(worldStartLevels, 0, tmp, 0, worldStartLevels.length);
    worldStartLevels = tmp;
    maxWorlds = newWorldCapacity;
 }
  /**
   * Moving up to the next world.
   */

  public void worldPush() {
    this.worldStartLevels[this.environment.currentWorld + 1] = currentLevel;
  }


  /**
   * Moving down to the previous world.
   */

  public void worldPop() {
    while (currentLevel > worldStartLevels[this.environment.currentWorld]) {
      currentLevel--;
      StoredIntVector v = vectorStack[currentLevel];
      v._set(indexStack[currentLevel], valueStack[currentLevel], stampStack[currentLevel]);
    }
  }


  /**
   * Comits a world: merging it with the previous one.
   */

  public void worldCommit() {
    // TODO
  }


  /**
   * Returns the current size of the stack.
   */

  public int getSize() {
    return currentLevel;
  }
}
