//     ___  ___         PaLM library for Choco system
//    /__ \/ __\        (c) 2001 - 2004 -- Narendra Jussien
//   ____\ |/_____
//  /_____\/_____ \     Explanation based constraint programming tool
// |/   (_)(_)   \|
//       \ /            Version 0.1
//       \ /            January 2004
//       \ /
// ______\_/_______     Contibutors: Franois Laburthe, Hadrien Cambazard, Guillaume Rochart...

package choco.palm;

import choco.palm.integer.PalmIntVar;
import choco.palm.integer.constraints.*;
import choco.palm.prop.PalmEngine;
import choco.palm.search.PalmGlobalSearchSolver;
import choco.palm.search.PalmContradiction;
import choco.palm.explain.Explanation;
import choco.palm.explain.PalmConstraintPlugin;
import choco.palm.explain.GenericExplanation;
import choco.palm.real.PalmRealVarImpl;

import java.util.List;
import java.util.Iterator;
import java.util.BitSet;
import java.util.logging.*;

import choco.prop.ChocEngine;
import choco.prop.ConstraintEvent;
import choco.integer.IntVar;
import choco.integer.IntExp;
import choco.integer.var.IntDomainVar;
import choco.integer.var.IntTerm;
import choco.integer.constraints.IntLinComb;
import choco.*;
import choco.real.RealVar;
import choco.real.var.RealVarImpl;

/**
 * Choco problem extension involving explanations and explanation-based algorithms (mac-dbt, decision-repair...)
 */

public class PalmProblem extends Problem {
  /**
   * Constant value for relaxation number statistics.
   */

  public final static int RLX = 0;


  /**
   * Constant value for extension number statistics.
   */

  public final static int EXT = 1;


  /**
   * Constant value for CPU statistics.
   */

  public final static int CPU = 2;


  /**
   * States if the release information should be displayed during the next problem instantiation.
   */

  public static boolean displayRelease = true;


  /**
   * Maximum relaxation level acceptable without user interaction.
   */

  public int maxRelaxLevel = 0;


  /**
   * The logger used in Palm (The name is <code>choco.palm</code>).
   */

  protected Logger logger;


  /**
   * Set with last erased constraints (index that can be used for posted constraints)
   */

  protected BitSet erasedCst;


  /**
   * Displays release information (date, verions, ...).
   */

  public final static void ReleaseDisplay() {
    System.out.println("** JPaLM : Constraint Programming with Explanations");
    System.out.println("** JPaLM v0.1 (September, 2003), Copyright (c) 2000-2003 N. Jussien");
    displayRelease = false;
  }


  /**
   * Creates a Palm Problem with the specified environment.
   */
  public PalmProblem() {
    super();

    // Ensures a determinist behaviour
    GenericExplanation.reinitTimestamp();
    erasedCst = new BitSet();

    // Specialized engine and solver for Palm
    this.propagationEngine = new PalmEngine(this);
    this.solver = new PalmSolver(this);

    // Displays information about Palm
    if (displayRelease) ReleaseDisplay();

    // Logger initialization
    logInit();
  }

  /**
   * Inits the palm Logger "choco.palm".
   */
  protected void logInit() {
    logger = Logger.getLogger("choco.palm");
    logger.setUseParentHandlers(true);
    logger.setLevel(Level.ALL);
  }

  /**
   * Factory to create explanation.
   * It offers the possibility to make another kind of explanation, only by extending PalmProblem 
   * @return the new explanation object
   */
  public Explanation makeExplanation() {
    return new GenericExplanation(this);
  }

  /**
   * Returns all variables of the variables.
   * @deprecated
   */
  public IntVar[] getVars() {
    IntVar[] array = new IntVar[0];
    return (IntVar[]) this.intVars.toArray(array);
  }


  /**
   * Makes a new Integer Variable with a bucket for managing holes in the domain.
   * @param name Name of the variable (for display).
   * @param inf Inf bound.
   * @param sup Sup bound.
   * @return The created variable.
   */
  public IntVar makeIntVar(String name, int inf, int sup) {
    return makeEnumIntVar(name, inf, sup);
  }

  /**
   * Makes a new Integer Variable with a bucket for managing holes in the domain.
   * @param name Name of the variable (for display).
   * @param inf Inf bound.
   * @param sup Sup bound.
   * @return The created variable.
   */
  public IntVar makeEnumIntVar(String name, int inf, int sup) {
    IntVar var = new PalmIntVar(this, name, IntVar.LIST, inf, sup);
    intVars.add(var);
    return var;
  }

  /**
   * Creates a new Bound Integer Variable without any bucket (only bounds are propagated).
   * @param name Name of the variable.
   * @param inf Inf bound.
   * @param sup Sup bound.
   * @return The created variable.
   */
  public IntVar makeBoundIntVar(String name, int inf, int sup) {
    return makeBoundIntVar(name, inf, sup, true);
  }


  /**
   * Creates a new Bound Integer Variable without any bucket (only bounds are propagated).
   * @param name Name of the variable.
   * @param inf Inf bound.
   * @param sup Sup bound.
   * @param toAdd Specifies if the variable should be added to the variable
   *  list of the problem (should be true in almost all cases).
   * @return The created variable.
   */
  public IntVar makeBoundIntVar(String name, int inf, int sup, boolean toAdd) {
    IntVar var = new PalmIntVar(this, name, IntVar.BOUNDS, inf, sup);
    if (toAdd) intVars.add(var);
    return var;
  }


  /**
   * Displays all the runtime statistics.
   */
  //TODO rename and add in Choco !!
  public void printRuntimeSatistics() {
      ((PalmGlobalSearchSolver)this.getSolver().getSearchSolver()).printRuntimeStatistics();
  }

  /**
   * Searches one solution of the problem.
   * @return True if a solution was found.
   * @deprecated
   */
  public Boolean searchOneSolution() {
    solver.getSearchSolver().run();
    return this.isFeasible();
  }


  /**
   * Tries to search the problem by finding one solution or all solutions.
   * @param allSolutions If true, all the solutions are searched.
   * @deprecated
   */
  public Boolean solve(boolean allSolutions) {
    solver.setFirstSolution(!allSolutions);
    solver.generateSearchSolver(this);
    try {
      solver.getSearchSolver().newTreeSearch();
    } catch (ContradictionException e) {
      System.out.println("Should not happen : PalmProblem.solve(boolean)");
    }
    if (allSolutions) {
      boolean soluble = (this.searchOneSolution() == Boolean.TRUE);
      while (soluble) {
        logger.info("A solution was found.");
        soluble = ((PalmGlobalSearchSolver)solver.getSearchSolver()).getState().discardCurrentSolution()
                && (this.searchOneSolution() == Boolean.TRUE);
      }
    } else return this.searchOneSolution();
    solver.getSearchSolver().endTreeSearch();
    return Boolean.TRUE;
  }

  /**
   * Tries to search the <i>first</i> solution of the problem.
   */
  public Boolean solve() {
    solver.generateSearchSolver(this);
    try {
      solver.getSearchSolver().newTreeSearch();
    } catch (ContradictionException e) {
      System.out.println("Should not happen : PalmProblem.solve(boolean)");
    }
    solver.getSearchSolver().run();
    if (this.isFeasible() == Boolean.FALSE) {
      solver.getSearchSolver().endTreeSearch();
    }
    return this.isFeasible();
  }

  /**
   * Tries to find another solution.
   */
  public Boolean nextSolution() {
    if (((PalmGlobalSearchSolver)solver.getSearchSolver()).getState().discardCurrentSolution()) {
      solver.getSearchSolver().run();
    } else {
      solver.getSearchSolver().endTreeSearch();
      return Boolean.FALSE;
    }
    if (this.isFeasible() == Boolean.FALSE) {
      solver.getSearchSolver().endTreeSearch();
    }
    return this.isFeasible();
  }

  /**
   * Checks if current solution is still valid.
   */
  public boolean checkSolution() {
    try {
      this.propagate();
      return true;
    } catch(PalmContradiction e) {
      return false;
    } catch(ContradictionException e) {
      logger.severe("This should not happen: PalmProblem.checkSolution()");
      return false;
    }
  }

  /**
   * Tries to find all solutions of the problem.
   */
  public boolean solveAll() {
    solver.generateSearchSolver(this);
    try {
      solver.getSearchSolver().newTreeSearch();
    } catch (ContradictionException e) {
      System.out.println("Should not happen : PalmProblem.solve(boolean)");
    }
    solver.getSearchSolver().run();
    boolean soluble = (this.isFeasible() == Boolean.TRUE);
    while (soluble) {
      logger.info("A solution was found.");
      soluble = ((PalmGlobalSearchSolver)solver.getSearchSolver()).getState().discardCurrentSolution();
      if (soluble) {
        solver.getSearchSolver().run();
        soluble = (this.isFeasible() == Boolean.TRUE);
      }
    }
    solver.getSearchSolver().endTreeSearch();
    return (isFeasible() == Boolean.TRUE);
  }

  /**
   * Maximize an objective variable with a PalmBranchAndBound
   */
  public Boolean maximize(IntVar objective) {
    ((PalmSolver)solver).setObjective(objective);
    ((PalmSolver)solver).setMaximize();
    solver.generateSearchSolver(this);
    solver.getSearchSolver().run();
    return this.isFeasible();
  }


  /**
   * Minimize an objective variable with a PalmBranchAndBound
   */
  public Boolean minimize(IntVar objective) {
    ((PalmSolver)solver).setObjective(objective);
    ((PalmSolver)solver).setMinimize();
    solver.generateSearchSolver(this);
    solver.getSearchSolver().run();
    return this.isFeasible();
  }

  /**
   * Posts a constraints in the problem. If it has ever been posted (but deactivated), it is
   * only reactivated and repropagated.
   * @param constraint The constraint to post.
   */
  public void post(Constraint constraint) {
    if (!((PalmConstraintPlugin) constraint.getPlugIn()).isEverConnected()) {
      int idx;
      if (erasedCst.length() == 0) {
        constraints.add(constraint);
        idx = this.constraints.size() - 1;
      } else {
        idx = erasedCst.nextSetBit(0);
        constraints.set(idx, constraint);
        erasedCst.clear(idx);
      }
      ((PalmConstraintPlugin) constraint.getPlugIn()).setConstraintIdx(idx);
      constraint.addListener();
      ConstraintEvent event = (ConstraintEvent) constraint.getEvent();
      ((ChocEngine) this.propagationEngine).registerEvent(event);
      ((ChocEngine) this.propagationEngine).postConstAwake(constraint, true);
    } else {
      logger.fine("The constraint " + constraint + " is reactivated.");
      ((ChocEngine) this.propagationEngine).postConstAwake(constraint, true);
      constraint.activateListener();
    }
    if (((PalmConstraintPlugin) constraint.getPlugIn()).isDepending())
      ((PalmConstraintPlugin) constraint.getPlugIn()).setDependance();
  }


  /**
   * Posts a constraint with the specified weight.
   * @param constraint The constraint to post.
   * @param w The weight associated to the constraint.
   */
  public void post(Constraint constraint, int w) {
    this.post(constraint);
    ((PalmConstraintPlugin) constraint.getPlugIn()).setWeight(w);
  }


  /**
   * Posts an indirect constraint with an explain.
   * @param constraint The constraint to post.
   * @param expl The set of constraint this posted constraint depends on.
   */
  public void post(Constraint constraint, Explanation expl) {
    ((PalmConstraintPlugin) constraint.getPlugIn()).setIndirect(expl);
    this.post(constraint);
  }

  public void eraseConstraint(int constraintIdx) {
    constraints.set(constraintIdx, null);
    erasedCst.set(constraintIdx);
  }


  /**
   * Posts and propagates several decision constraints (that is decisions taken by the solver).
   * @param constraints The constraints to post.
   * @throws ContradictionException
   */
  public void propagateAllDecisionsConstraints(List constraints) throws ContradictionException {
    //this.palmSolver.incRuntimeStatistic(EXT, 1);  move in explore
    for (Iterator iterator = constraints.iterator(); iterator.hasNext();) {
      Constraint constraint = (Constraint) iterator.next();
      this.post(constraint, 0); // Avant la mise a jour de l'etat sinon la contrainte n'existe pas encore !!
      ((PalmGlobalSearchSolver)solver.getSearchSolver()).getState().addDecision(constraint);
      this.propagate();
    }
  }


  /**
   * Tries to repair the problem after a PalmContradiction thanks to removing a responsible
   * constraint (that is a constraint in the explain of the contradiction).
   * @throws ContradictionException
   */
  public void repair() throws ContradictionException {
    ((PalmGlobalSearchSolver)solver.getSearchSolver()).repair();
  }


  /**
   * Removes properly a constraint from the problem: the constraint is deactivated and all the depending
   * filtering decisions are undone.
   * @param constraint The constraint to remove.
   */
  public void remove(Constraint constraint) {
    ((PalmEngine) this.propagationEngine).remove(constraint);
    if (((PalmConstraintPlugin)constraint.getPlugIn()).isEphemeral()) {
      //System.out.println("+Remove : " + ((PalmConstraintPlugin)constraint.getPlugIn()).getConstraintIdx());
      //this.eraseConstraint(((PalmConstraintPlugin)constraint.getPlugIn()).getConstraintIdx());
      ((AbstractConstraint) constraint).delete();
    }
  }


  /**
   * Gets the minimal domain variable of the problem. This is used for choosing a variable to instantiate
   * by the solver.
   * @return The minimal domain variable.
   */
  public Var getMinDomVar() {
    int minsize = Integer.MAX_VALUE;
    Var var = null;
    for (int i = 0; i < intVars.size(); i++) {
      PalmIntVar v = (PalmIntVar) intVars.get(i);
      int size = v.getDomainSize();
      if (size > 1) {
        if (size < minsize) {
          minsize = size;
          var = v;
        }
      }
    }
    return var;
  }

  /**
   * @param nb Constraint number (number affected when posting and stored
   * in the variable plugin)
   * @return Returns the constraint
   */
  public Constraint getConstraintNb(int nb) {
    return (Constraint) this.constraints.get(nb);
  }

  /**
   * Specialized method for Palm unary inequality constraint
   * @param x An integer expression
   * @param c A value the expression cannot equal to
   * @return the constraint ensuring such inequality
   */
  public Constraint neq(IntExp x, int c) {
    if (x instanceof IntVar) {
      return new PalmNotEqualXC((IntVar) x, c);
    } else if (x instanceof IntTerm) {
      IntTerm t = (IntTerm) x;
      if ((t.getSize() == 2) && (t.getCoefficient(0) + t.getCoefficient(1) == 0)) {
        return new PalmNotEqualXYC(t.getVariable(0), t.getVariable(1), (c - t.getConstant()) / t.getCoefficient(0));
      } else {
        return new PalmIntLinComb(((IntTerm) x).getVariables(), ((IntTerm) x).getCoefficients(), -(c), IntLinComb.NEQ);
      }
    } else {
      throw new Error("IntExp not a good exp");
    }
  }

  /**
   * Specialized method for Palm binary inequality constraint
   * @param x An integer expression
   * @param y An integer expression
   * @return the constraint ensuring such inequality
   */
  public Constraint neq(IntExp x, IntExp y) {
    if (x instanceof IntVar && y instanceof IntVar) {
      return new PalmNotEqualXYC((IntVar) x, (IntVar) y, 0);
    } else if ((x instanceof IntTerm || x instanceof IntVar) && (y instanceof IntTerm || y instanceof IntVar)) {
      return neq(this.minus(x, y), 0);
    } else {
      throw new Error("IntExp not a good exp");
    }
  }

  /**
   * Specialized method for Palm unary superiority constraint
   * @param x An integer expression
   * @param c A value that expression can not be instantiated to
   * @return the constraint ensuring such superiority
   */
  public Constraint geq(IntExp x, int c) {
    if (x instanceof IntVar) {
      return new PalmGreaterOrEqualXC((IntVar) x, c);
    } else if (x instanceof IntTerm) {
      IntTerm t = (IntTerm) x;
      if ((t.getSize() == 2) && (t.getCoefficient(0) + t.getCoefficient(1) == 0)) {
        return new PalmGreaterOrEqualXYC(t.getVariable(0), t.getVariable(1), (c - t.getConstant()) / t.getCoefficient(0));
      } else {
        return new PalmIntLinComb(((IntTerm) x).getVariables(), ((IntTerm) x).getCoefficients(), -(c), IntLinComb.GEQ);
      }
    } else {
      throw new Error("IntExp not a good exp");
    }
  }

  /**
   * Specialized method for Palm unary superiority constraint
   * @param x An integer expression
   * @param c A value that expression can not be instantiated to
   * @return the constraint ensuring such superiority
   */
  public Constraint geq(int c, IntExp x) {
    if (x instanceof IntVar) {
      return new PalmLessOrEqualXC((IntDomainVar) x, c);
    } else if (x instanceof IntTerm) {
      IntTerm t = (IntTerm) x;
      if ((t.getSize() == 2) && (t.getCoefficient(0) + t.getCoefficient(1) == 0)) {
        return new PalmGreaterOrEqualXYC(t.getVariable(1), t.getVariable(0), (t.getConstant() - c) / t.getCoefficient(0));
      } else {
        // TODO : mettre des moins partout
        throw new Error("IntExp not a good exp");
      }
    } else {
      throw new Error("IntExp not a good exp");
    }
  }

  /**
   * Specialized method for Palm binary superiority constraint
   * @param x An integer expression
   * @param y An integer expression
   * @return the constraint ensuring such superiority
   */
  public Constraint geq(IntExp x, IntExp y) {
    if (x instanceof IntVar && y instanceof IntVar) {
      return new PalmGreaterOrEqualXYC((IntVar) x, (IntVar) y, 0);
    } else if ((x instanceof IntTerm || x instanceof IntVar) && (y instanceof IntTerm || y instanceof IntVar)) {
      return geq(this.minus(x, y), 0);
    } else {
      throw new Error("IntExp not a good exp");
    }
  }

  /**
   * Specialized method for Palm unary equality constraint
   * @param x An integer expression
   * @param c An integer value the expression should equal
   * @return the constraint ensuring the equality
   */
  public Constraint eq(IntExp x, int c) {
    if (x instanceof IntVar) {
      return new PalmEqualXC((IntVar) x, c);
    } else if (x instanceof IntTerm) {
      IntTerm t = (IntTerm) x;
      if ((t.getSize() == 2) && (t.getCoefficient(0) + t.getCoefficient(1) == 0)) {
        return new PalmEqualXYC(t.getVariable(0), t.getVariable(1), (c - t.getConstant()) / t.getCoefficient(0));
      } else {
        return new PalmIntLinComb(((IntTerm) x).getVariables(), ((IntTerm) x).getCoefficients(), -(c), IntLinComb.EQ);
      }
    } else {
      throw new Error("IntExp not a good exp");
    }
  }

  /**
   * Specialized method for Palm binary equality constraint
   * @param x An integer expression
   * @param y An integer expression
   * @return the constraint ensuring the equality
   */
  public Constraint eq(IntExp x, IntExp y) {
    if (x instanceof IntVar && y instanceof IntVar) {
      return new PalmEqualXYC((IntVar) x, (IntVar) y, 0);
    } else if ((x instanceof IntTerm || x instanceof IntVar) && (y instanceof IntTerm || y instanceof IntVar)) {
      return eq(this.minus(x, y), 0);
    } else {
      throw new Error("IntExp not a good exp");
    }
  }

  /**
   * Returns the maximum level the solver can relax without user interaction.
   */
  public int getMaxRelaxLevel() {
    return maxRelaxLevel;
  }

  /**
   * Sets the maximum level the solver can relax without user interaction (default value is 0, that is only
   * decision constraints).
   * @param maxRelaxLevel the new level
   */
  public void setMaxRelaxLevel(int maxRelaxLevel) {
    this.maxRelaxLevel = maxRelaxLevel;
  }

  /**
   * Creates a real variable.
   * @param name
   * @param min
   * @param max
   * @return the new object (RealVar)
   */
  public RealVar makeRealVar(String name, double min, double max) {
    RealVar v = new PalmRealVarImpl(this, name, min, max);
    floatVars.add(v);
    return v;
  }
}
