// *********************************************
// *                   J-CHOCO                 *
// *   Copyright (c) F. Laburthe, 1999-2003    *
// *********************************************
// * Event-base contraint programming Engine   *
// *********************************************

// CVS Information
// File:               $RCSfile: VarEvent.java,v $
// Version:            $Revision: 1.1.1.1 $
// Last Modification:  $Date: 2004/05/11 13:51:50 $
// Last Contributor:   $Author: laburthe $

package choco.prop;

import choco.AbstractVar;
import choco.ContradictionException;
import choco.Problem;
import choco.integer.var.IntDomainVar;
import choco.mem.StoredIntVector;
import choco.util.IPrioritizable;
import choco.util.IntIterator;
import choco.util.StoredPointerCycle;

import java.util.logging.Logger;

/**
 * Implements an
 * {@link choco.prop.PropagationEvent} for the variable events.
 */
public abstract class VarEvent implements PropagationEvent, IPrioritizable {

  /**
   * empty bitvector for the event type.
   */
  public final static int NOEVENT = -2;

  /**
   * Semantic of the cause of an event: -1 means that the event is active without
   * any precise cause. (Equivalent of 0 in Claire version)
   */
  public final static int NOCAUSE = -1;

  /**
   * Cause of this basic var.
   */
  protected int cause = NOEVENT;

  /**
   * empty bitvector for the event type.
   */
  public final static int EMPTYEVENT = 0;

  /**
   * The touched variable.
   */

  protected AbstractVar modifiedVar;


  /**
   * A linked list with all the constraint to propagate.
   */

  protected StoredIntVector nextConst;

  /**
   * Reference to the root Logger, an object for logging trace statements related to propagation events (using the java.util.logging package)
   */

  protected static Logger logger = Logger.getLogger("choco.prop");

  /**
   * A pointer cycle encoding the set of constraints connected to the event
   */
  protected StoredPointerCycle activeCycle;

  /**
   * stores the type of update performed on the variable
   */
  protected int eventType = EMPTYEVENT;

  /**
   * Constructs a variable event for the specified variable and with the given
   * basic events.
   */

  public VarEvent(AbstractVar var) {
    this.modifiedVar = var;
    activeCycle = new StoredPointerCycle(getProblem().getEnvironment());
  }

  public int getNbListeners() {
    return activeCycle.size();
  }

  /**
   * Returns the touched variable.
   */

  public AbstractVar getModifiedVar() {
    return modifiedVar;
  }

  /**
   * Returns the touched variable.
   */

  public Object getModifiedObject() {
    return modifiedVar;
  }

  /**
   * freezes the state of the "delta domain": the set of values that are considered for removal
   * from the domain. Further removals will be treated as a further event.
   */
  protected void freeze() {
    cause = NOEVENT;
  }

  protected boolean release() {
    boolean anyUpdateSinceFreeze = (cause != NOEVENT);
    return anyUpdateSinceFreeze;
  }

  /**
   * Propagates the event through calls to the propagation engine.
   * @return true if the event has been fully propagated (and can thus be discarded), false otherwise
   * @throws ContradictionException
   */
  public abstract boolean propagateEvent() throws ContradictionException;

  /**
   * Clears the var: delegates to the basic events.
   */
  public abstract void clear();

  /**
   * Checks if a given listener is active or not
   * @param idx the index of the listener among all listeners connected to the variable
   */

  public boolean isActive(int idx) {
    return activeCycle.isInCycle(idx);
  }

  /**
   * Retrieving the problem.
   */

  public Problem getProblem() {
    return modifiedVar.getProblem();
  }

  /**
   * Adds the constraint <code>idx</code> as a listener.
   * @param idx the index of the listener among all listeners connected to the variable
   * @param active  Specifies if the listener should be active.
   */
  public void addListener(int idx, boolean active) {   // ex connectEvent
    activeCycle.add(idx, active);
  }

  /**
   * Activates the listeners.
   * @param idx the index of the listener among all listeners connected to the variable
   */
  public void activateListener(int idx) { // ex disconnectEvent
    activeCycle.setInCycle(idx);
  }

  /**
   * Desactivates the listeners.
   * @param idx the index of the listener among all listeners connected to the variable
   */

  public void deactivateListener(int idx) { // ex reconnectEvent
    activeCycle.setOutOfCycle(idx);
  }

  /**
   * Returns the cause of this basic var.
   */
  public int getCause() {
    return cause;
  }

  public int getPriority() {
    return 1;
  }

  public IntIterator getAllListenersButIterator(int avoidIndex) {
    return activeCycle.getCycleButIterator(avoidIndex);
  }

  /**
   * tests whether the event is currently active (present in some queue) or not
   * @return true if and only if the event is present in some queue, waiting to be handled
   * (returns false if the event is either absent from the queue or is the current event,
   * just popped from the queue and being propagated)
   */
  public boolean isEnqueued() {
    return (eventType != EMPTYEVENT);
  }

  public void recordEventTypeAndCause(int basicEvt, int idx) {
    // if no such event was active on the same variable
    if ( (cause == NOEVENT) || (eventType == EMPTYEVENT)) {  // note: these two tests should be equivalent
      // the varevent is reduced to basicEvt, and the cause is recorded
      eventType = (1 << basicEvt);
      cause = idx;
    } else {
      // otherwise, this basic event is added to all previous updates that are possibly mending on the same variable
      eventType = (eventType | (1 << basicEvt));
      // in case the cause of this update is different from the previous cause, all causes are forgotten
      // (so that the constraints that caused the event will be reawaken)
      if (cause != idx) {
        cause = NOCAUSE;
      }
    }
  }

}
